import textwrap
from decimal import Decimal


def construct_invoice_json(invoice):
    NS = 'http://www.w3.org/2001/XMLSchema-instance'

    data = {
        '_name': 'Invoice',
        '_attrs': [
            ('xmlns', 'urn:eslog:2.00'),
            ("{%s}noNamespaceSchemaLocation" % NS, 'http://www.roseslovenia.eu/e_files/news/eSLOG20_INVOIC_v200.xsd'),
        ],
        '_ns': {
            'xs4xs': 'http://www.w3.org/2001/XMLSchema',
            'in': 'http://uri.etsi.org/01903/v1.1.1#',
            'ds': 'http://www.w3.org/2000/09/xmldsig#',
            'xsi': NS
        },
        'invoice': {
            '_name': 'M_INVOIC',
            '_attrs': [('Id', 'data')],
            '_sorting': ['S_UNH', 'S_BGM', 'S_DTM', 'S_PAI', 'S_ALI', 'S_FTX', 'G_SG1', 'G_SG2', 'G_SG6', 'G_SG7',
                         'G_SG8', 'G_SG9', 'G_SG12', 'G_SG14', 'G_SG16', 'G_SG24', 'G_SG26', 'S_UNS', 'S_CNT', 'G_SG50',
                         'G_SG52', 'G_SG53', 'S_UNT'],
            'document_header': construct_document_header_data(invoice),
            'header': construct_header_data(invoice),
            'date_issued': construct_date_data(invoice.date_issued_code, invoice.date_issued),
            'date_of_service': construct_date_data(invoice.date_of_service_code, invoice.date_of_service),
            'payment_type': construct_payment_type_data(invoice),
            'payment_purpose': construct_payment_purpose_data(invoice),
            'issuer': construct_company_data(invoice.issuer, 'SE'),
            'buyer': construct_company_data(invoice.recipient, 'BY'),
            # 'recipient': construct_company_data(invoice.recipient, 'IV'),
            'currency': construct_currency_data(invoice.currency),
            'payment_terms': construct_payment_terms_data(invoice.date_due_code, invoice.date_due),
            'payment_data': construct_payment_data(invoice.total_with_tax)
        }  # end invoice
    }

    if invoice.payment_reference != None:
        data['invoice']['payment_reference_data'] = construct_payment_reference_data(invoice.payment_reference)

    for i, reference_document in enumerate(invoice.reference_documents):
        data['invoice'][f"reference_document_{i}"] = construct_reference_document_data(reference_document)

    if invoice.global_discount_amount:
        data['invoice']['global_discount'] = construct_global_discount_data(invoice.global_discount_amount, invoice.global_discount_percentage)

    if invoice.intro_text:
        data['invoice']['intro_text'] = construct_custom_text_data('GEN', invoice.intro_text)

    for i, item in enumerate(invoice.document_items):
        data['invoice'][f"item_{i}"] = construct_item_data(item)

    for i, ts in enumerate(invoice.tax_summaries):
        data['invoice'][f"tax_summary_{i}"] = construct_tax_summary_data(ts)

    # add final sums to invoice
    # Total without discount
    data['invoice']['sums_without_discounts'] = construct_sums_data(amount=invoice.total_without_discount, sum_type='79')
    # Discounts amount
    data['invoice']['sums_discounts'] = construct_sums_data(amount=invoice.total_without_discount - invoice.total_without_tax, sum_type='260')
    # Tax base sums
    data['invoice']['sums_tax_base_amount'] = construct_sums_data(amount=invoice.total_without_tax, sum_type='389')
    # Taxes amount
    data['invoice']['sums_taxes'] = construct_sums_data(amount=invoice.total_with_tax - invoice.total_without_tax, sum_type='176')
    # Total amount - with taxes
    data['invoice']['sums_total_amount'] = construct_sums_data(amount=invoice.total_with_tax, sum_type='388')

    if invoice.outro_text:
        data['invoice']['outro_text'] = construct_custom_text_data('GEN', invoice.outro_text)

    return data


def construct_document_header_data(invoice):
    header = {
        '_name': 'S_UNH',
        'invoice_no': {
            '_name': 'D_0062',
            '_value': invoice.invoice_number[-14:]  # last 14 chars if longer...else whole string
        },
        'data': {
            '_name': 'C_S009',
            'val1': {
                '_name': 'D_0065',
                '_value': 'INVOIC'
            },
            'val2': {
                '_name': 'D_0052',
                '_value': 'D'
            },
            'val3': {
                '_name': 'D_0054',
                '_value': '01B'
            },
            'val4': {
                '_name': 'D_0051',
                '_value': 'UN'
            }
        }

    }

    return header


def construct_header_data(invoice):
    header = {
        '_name': 'S_BGM',
        'invoice_type': {
            '_name': 'C_C002',
            'value_tag': {
                '_name': 'D_1001',
                '_value': invoice.invoice_type,
            }
        },
        'invoice_number': {
            '_name': 'C_C106',
            'value_tag': {
                '_name': 'D_1004',
                '_value': invoice.invoice_number,
            }
        }
    }

    return header


def construct_payment_type_data(invoice):
    header = {
        '_name': 'S_FTX',
        'key': {
            '_name': 'D_4451',
            '_value': 'PAI'
        },
        'value': {
            '_name': 'C_C108',
            'value_tag': {
                '_name': 'D_4440',
                '_value': invoice.payment_type,
            }
        }
    }

    return header


def construct_payment_purpose_data(invoice):
    header = {
        '_name': 'S_FTX',
        'key': {
            '_name': 'D_4451',
            '_value': 'ALQ'

        },
        'value': {
            '_name': 'C_C108',
            'value_tag': {
                '_name': 'D_4440',
                '_value': invoice.payment_purpose,
            }
        }
    }

    return header


def construct_date_data(date_code, date):
    date = {
        '_name': 'S_DTM',
        'wrapper': {
            '_name': "C_C507",
            'date_type': {
                '_name': 'D_2005',
                '_value': date_code
            },
            'date': {
                '_name': 'D_2380',
                '_value': date.strftime('%Y-%m-%d')
            }
        }
    }

    return date


def construct_currency_data(currency):
    currency = {
        '_name': 'G_SG7',
        'wrapper': {
            '_name': 'S_CUX',
            'wrapper': {
                '_name': 'C_C504',
                'currency_type': {
                    '_name': 'D_6347',
                    '_value': '2',
                },
                'currency_code': {
                    '_name': 'D_6345',
                    '_value': currency,
                },
            }
        }
    }

    return currency


def construct_location_data(location_code, location_address):  # does not translate
    location = {
        '_name': 'Lokacije',
        'location_code': {
            '_name': 'VrstaLokacije',
            '_value': location_code,
        },
        'location_name': {
            '_name': 'NazivLokacije',
            '_value': location_address,
        },
    }

    return location


def construct_company_data(business, business_type='SE'):
    data = {
        '_name': 'G_SG2',
        'info1': {
            '_name': 'S_NAD',
            'business_type': {
                '_name': 'D_3035',
                '_value': business_type
            },
            'name': {
                '_name': 'C_C080',
            },
            'address': {
                '_name': 'C_C059',
            },
            'city': {
                '_name': 'D_3164',
                '_value': business.city
            },
            'zip_code': {
                '_name': 'D_3251',
                '_value': str(business.zip_code)
            },
            'country_code': {
                '_name': 'D_3207',
                '_value': business.country_iso_code
            }
        }
    }

    # TODO: check if ok
    # Add business name
    business_name_split = textwrap.wrap(business.name, 70, break_long_words=True)

    for i, bn_part in enumerate(business_name_split):
        i = i + 1  # Start from 1
        data['info1']['name'][f"part_{i}"] = {
            '_name': f"D_3036_{i}",
            '_value': bn_part
        }

        if i == 1:
            data['info1']['name'][f"part_{i}"]["_name"] = "D_3036"

        if i == 4:
            break  # Stop at max length

    # Add business address
    addr_split = textwrap.wrap(business.address, 35, break_long_words=True)

    for i, addr_part in enumerate(addr_split):
        i = i + 1  # Start from 1
        data['info1']['address'][f"part_{i}"] = {
            '_name': f"D_3042_{i}",
            '_value': addr_part
        }

        if i == 1:
            data['info1']['address'][f"part_{i}"]["_name"] = "D_3042"

        if i == 4:
            break  # Stop at max length

    if business.iban:
        data['financial_info'] = {
            '_name': 'S_FII',
            'type': {
                '_name': 'D_3035',
                '_value': 'RB' if business_type == 'SE' else 'BB'  # RB - Receiving Bank or BB - Buyer Bank
            },
            'bank_account_info1': {
                '_name': 'C_C078',
                'iban': {
                    '_name': 'D_3194',
                    '_value': business.iban
                },
                'owner': {
                    '_name': 'D_3192',
                    '_value': business.name[:34]
                }
            },
            'bank_account_info2': {
                '_name': 'C_C088',
                'bic': {
                    '_name': 'D_3433',
                    '_value': business.bic
                }
            }
        }

    if business.vat_id:
        data['vat_id'] = {
            '_name': 'G_SG3',
            'wrapper': {
                '_name': 'S_RFF',
                'wrapper': {
                    '_name': 'C_C506',
                    'type': {
                        '_name': 'D_1153',
                        '_value': 'VA'
                    },
                    'vat': {
                        '_name': 'D_1154',
                        '_value': str(business.vat_id)
                    }
                }
            }
        }

    if business.registration_number:
        data['registration_number'] = {
            '_name': 'G_SG3',
            'wrapper': {
                '_name': 'S_RFF',
                'wrapper': {
                    '_name': 'C_C506',
                    'type': {
                        '_name': 'D_1153',
                        '_value': '0199'
                    },
                    'vat': {
                        '_name': 'D_1154',
                        '_value': str(business.registration_number)
                    }
                }
            }
        }

    return data


def construct_payment_terms_data(date_due_code, date_due):
    payment_terms = {
        '_name': 'G_SG8',
        'term_data': {
            '_name': 'S_PAT',
            'term_code': {
                '_name': 'D_4279',
                '_value': '1',
            }
        },
        'term_due': {
            '_name': 'S_DTM',
            'wrapper': {
                '_name': 'C_C507',
                'term_due_type': {
                    '_name': 'D_2005',
                    '_value': date_due_code,
                },
                'term_due_date': {
                    '_name': 'D_2380',
                    '_value': date_due.strftime('%Y-%m-%d'),
                },
            }
        }
    }

    return payment_terms


def construct_payment_data(total_with_tax):
    reference = {
        '_name': 'G_SG50',
        'invoice_amounts': {
            '_name': 'S_MOA',
            'wrapper': {
                '_name': 'C_C516',
                'type': {
                    '_name': 'D_5025',
                    '_value': '9'  # Amount to be paid
                },
                'amount': {
                    '_name': 'D_5004',
                    '_value': str(total_with_tax)
                }
            }
        }
    }

    return reference


def construct_payment_reference_data(payment_reference):
    reference = {
        '_name': 'G_SG1',
        'reference': {
            '_name': 'S_RFF',
            'wrapper': {
                '_name': 'C_C506',
                'ref_type': {
                    '_name': 'D_1153',
                    '_value': 'PQ'
                },
                'ref_number': {
                    '_name': 'D_1154',
                    '_value': payment_reference
                }
            }

        }
    }

    return reference


def construct_reference_document_data(reference_document):
    reference_doc_data = {
        '_name': 'G_SG1',
        'wrapper': {
            '_name': 'S_RFF',
            'wrapper': {
                '_name': 'C_C506',
                'document_type': {
                    '_name': 'D_1153',
                    '_value': reference_document.type_code
                },
                'document_number': {
                    '_name': 'D_1154',
                    '_value': reference_document.document_number
                }
            }
        }
    }

    return reference_doc_data


def construct_global_discount_data(discount_amount, discount_percentage):
    discount_data = {
        '_name': 'G_SG16',
        'description': {
            '_name': 'S_ALC',
            'type': {
                '_name': 'D_5463',
                '_value': 'A'
            },
            'wrapper': {
                '_name': 'C_C552',
                'desc': {
                    '_name': 'D_1230',
                    '_value': 'SKUPNI POPUST'
                },
                'type': {
                    '_name': 'D_5189',
                    '_value': '42'  # check if correct
                }
            }
        },
        'percentage': {
            '_name': 'G_SG19',
            'wrapper': {
                '_name': 'S_PCD',
                'wrapper': {
                    '_name': 'C_C501',
                    'type': {
                        '_name': 'D_5245',
                        '_value': '1'
                    },
                    'val': {
                        '_name': 'D_5482',
                        '_value': str(discount_percentage)
                    }
                }
            }
        },
        'amount': {
            '_name': 'G_SG20',
            'wrapper': {
                '_name': 'S_MOA',
                'wrapper': {
                    '_name': 'C_C516',
                    'type': {
                        '_name': 'D_5025',
                        '_value': '1'
                    },
                    'val': {
                        '_name': 'D_5004',
                        '_value': str(discount_amount)
                    }
                }
            }
        },
    }

    return discount_data


def construct_custom_text_data(text_format, text):
    """
    Text must be split into 70 chars long strings.
    :param text_format: AAI or other predefined formats
    :param text_type:
    :param text:
    :return:
    """
    text_split = textwrap.wrap(text, 70, break_long_words=True)

    custom_text = {
        '_name': 'S_FTX',
        'format': {
            '_name': 'D_4451',
            '_value': text_format,
        },
        'content': {
            '_name': 'C_C108'
        }
    }

    for i, txt in enumerate(text_split):
        # Since text_1 is used for text_type we must enumerate from 2 onwards
        i = i + 1

        name = 'D_4440'
        if i > 1:
            name = f"D_4440_{i}"

        custom_text['content'][f"text_{i}"] = {
            '_name': name,
            '_value': txt
        }

        # Stop the loop if index is 5 - we can't place any more than  that in XML.
        if i == 5:
            break

    return custom_text


def construct_item_data(item):
    data = {
        '_name': 'G_SG26',
        '_sorting': ['S_LIN', 'S_PIA', 'S_IMD', 'S_MEA', 'S_QTY', 'S_ALI', 'S_DTM', 'S_GIN', 'S_QVR', 'S_FTX', 'G_SG27',
                     'G_SG28', 'G_SG29', 'G_SG31', 'G_SG33', 'G_SG34', 'G_SG35', 'G_SG39', 'G_SG45', 'G_SG47'],
        'info': {
            '_name': 'S_LIN',
            'row_num': {
                '_name': 'D_1082',
                '_value': str(item.row_number)
            }
        },
        'description': {
            '_name': 'S_IMD',
            'code': {
                '_name': 'D_7077',
                '_value': item.item_description_code,
            },
            'name': {
                '_name': 'C_C273',
                'desc_1': {
                    '_name': 'D_7008',
                    '_value': item.item_name[:35]  # Only 35 chars...
                }
            }
        },
        'quantity': {
            '_name': 'S_QTY',
            'wrapper': {
                '_name': 'C_C186',
                'qty_type': {
                    '_name': 'D_6063',
                    '_value': item.quantity_type,
                },
                'qty': {
                    '_name': 'D_6060',
                    '_value': str(item.quantity)
                },
                'unit': {
                    '_name': 'D_6411',
                    '_value': item.unit
                }
            }
        },
        'value_before_discount': {
            '_name': 'G_SG27',
            'wrapper': {
                '_name': 'S_MOA',
                'wrapper': {
                    '_name': 'C_C516',
                    'type': {
                        '_name': 'D_5025',
                        '_value': '203'  # Total before discount
                    },
                    'amount': {
                        '_name': 'D_5004',
                        '_value': "%.2f" % (item.price_without_tax * item.quantity)
                    }
                }
            }
        },
        'value_total': {
            '_name': 'G_SG27',
            'wrapper': {
                '_name': 'S_MOA',
                'wrapper': {
                    '_name': 'C_C516',
                    'type': {
                        '_name': 'D_5025',
                        '_value': '38'  # Total before discount
                    },
                    'amount': {
                        '_name': 'D_5004',
                        '_value': str(item.total_with_tax)
                    }
                }
            }
        },
        'tax_info': {
            '_name': 'G_SG34',
            'taxes': {
                '_name': 'S_TAX',
                'id': {
                    '_name': 'D_5283',
                    '_value': '7'
                },
                'type': {
                    '_name': 'C_C241',
                    'value': {
                        '_name': 'D_5153',
                        '_value': 'VAT'
                    }
                },
                'vat_percentage': {
                    '_name': 'C_C243',
                    'value': {
                        '_name': 'D_5278',
                        '_value': str(item.tax_rate)
                    }
                },
                'tax_type': {
                    '_name': 'D_5305',
                    '_value': item.tax_rate_type
                },
            },
            'tax_amounts_base': {
                '_name': 'S_MOA',
                'wrapper': {
                    '_name': 'C_C516',
                    'type': {
                        '_name': 'D_5025',
                        '_value': '125'
                    },
                    'amount': {
                        '_name': 'D_5004',
                        '_value': str(item.total_without_tax)
                    }
                }
            },
            'tax_amounts_tax': {
                '_name': 'S_MOA',
                'wrapper': {
                    '_name': 'C_C516',
                    'type': {
                        '_name': 'D_5025',
                        '_value': '124'
                    },
                    'amount': {
                        '_name': 'D_5004',
                        '_value': str(item.total_with_tax - item.total_without_tax)
                    }
                }
            }
        }
    }

    if item.ean != None:
        data['info']['ean'] = {
            '_name': 'C_C212',
            'code': {
                '_name': 'D_7140',
                '_value': str(item.ean)
            },
            'type': {
                '_name': 'D_7143',
                '_value': '0160'  # type ean
            }
        }

    price_wo_tax = item.price_without_tax
    if item.discount_percentage:
        price_wo_tax = price_wo_tax * Decimal(1 - (item.discount_percentage / 100))

    data['price'] = {
        '_name': 'G_SG29',
        'wrapper': {
            '_name': 'S_PRI',
            'wrapper': {
                '_name': 'C_C509',
                'type': {
                    '_name': 'D_5125',
                    '_value': 'AAA'
                },
                'value': {
                    '_name': 'D_5118',
                    '_value': "%.2f" % price_wo_tax
                },
                'qty': {
                    '_name': 'D_5284',
                    '_value': '1'
                },
                'unit': {
                    '_name': 'D_6411',
                    '_value': 'C62'
                }
            }
        }
    }

    if item.discount_percentage:
        data['price_wo_discount'] = {
            '_name': 'G_SG29',
            'wrapper': {
                '_name': 'S_PRI',
                'wrapper': {
                    '_name': 'C_C509',
                    'type': {
                        '_name': 'D_5125',
                        '_value': 'AAB'
                    },
                    'value': {
                        '_name': 'D_5118',
                        '_value': str(item.price_without_tax)
                    },
                    'qty': {
                        '_name': 'D_5284',
                        '_value': '1'
                    },
                    'unit': {
                        '_name': 'D_6411',
                        '_value': 'C62'
                    }
                }
            }
        }
        data['discount'] = {
            '_name': 'G_SG39',
            'type': {
                '_name': 'S_ALC',
                'identification': {
                    '_name': 'D_5463',
                    '_value': 'A',  # Discount
                }
            },
            'percentage': {
                '_name': 'G_SG41',
                'wrapper': {
                    '_name': 'S_PCD',
                    'wrapper': {
                        '_name': 'C_C501',
                        'type': {
                            '_name': 'D_5245',
                            '_value': '1',  # Discount
                        },
                        'percentage': {
                            '_name': 'D_5482',
                            '_value': str(item.discount_percentage)
                        }
                    }
                }
            },
            'amount': {
                '_name': 'G_SG42',
                'wrapper': {
                    '_name': 'S_MOA',
                    'wrapper': {
                        '_name': 'C_C516',
                        'type': {
                            '_name': 'D_5025',
                            '_value': '204',  # Discount
                        },
                        'percentage': {
                            '_name': 'D_5004',
                            '_value': str(item.discount_amount)
                        }
                    }
                }
            }
        }

    return data


def construct_tax_summary_data(tax_summary):
    data = {
        '_name': 'G_SG52',
        'summary': {
            '_name': 'S_TAX',
            'type': {
                '_name': 'D_5283',
                '_value': '7'
            },
            'vat_wrapper': {
                '_name': 'C_C241',
                'type': {
                    '_name': 'D_5153',
                    '_value': 'VAT'
                }
            },
            'percentage_wrapper': {
                '_name': 'C_C243',
                'tax_percentage': {
                    '_name': 'D_5278',
                    '_value': str(tax_summary.tax_rate)
                }
            },
            'tax_type': {
                '_name': 'D_5305',
                '_value': tax_summary.tax_type
            }
        },
        'amount_base': {
            '_name': 'S_MOA',
            'wrapper': {
                '_name': 'C_C516',
                'type': {
                    '_name': 'D_5025',
                    '_value': '125'  # Osnova
                },
                'amount': {
                    '_name': 'D_5004',
                    '_value': str(tax_summary.tax_base)
                }
            }
        },
        'amount_tax': {
            '_name': 'S_MOA',
            'wrapper': {
                '_name': 'C_C516',
                'type': {
                    '_name': 'D_5025',
                    '_value': '124'  # Tax amount
                },
                'amount': {
                    '_name': 'D_5004',
                    '_value': str(tax_summary.tax_amount)
                }
            }
        }
    }

    return data


def construct_sums_data(amount, sum_type):
    data = {
        '_name': 'G_SG50',
        'amounts': {
            '_name': 'S_MOA',
            'wrapper': {
                '_name': 'C_C516',
                'type': {
                    '_name': 'D_5025',
                    '_value': str(sum_type)
                },
                'amount': {
                    '_name': 'D_5004',
                    '_value': "%.2f" % amount
                }
            }
        }
    }

    return data



from datetime import datetime, timezone


def convert_invoice_to_envelope(invoice, attachments, sender_bic, recipient_bic):
    NS = 'http://www.w3.org/2001/XMLSchema-instance'

    envelope = {
        '_name': 'envelope',
        '_ns': {
            'xsi': NS,
            'xsd': 'http://www.w3.org/2001/XMLSchema',
        },
        'sender': _construct_sender_receiver_data(invoice.issuer, 'sender', sender_bic),
        'receiver': _construct_sender_receiver_data(invoice.recipient, 'receiver', recipient_bic),
        'document_data': {
            '_name': 'doc_data',
            'type': {
                '_name': 'doc_type',
                '_value': '0002',
            },
            'version': {
                '_name': 'doc_type_ver',
                '_value': '01',
            },
            'document_id': {
                '_name': 'doc_id',
                '_value': str(invoice.invoice_number),
            },
            'external_document_id': {
                '_name': 'external_doc_id',
                '_value': str(invoice.invoice_number),
            },
            'timestamp': {
                '_name': 'timestamp',
                '_value': invoice.date_issued.isoformat()
            },
        },
        'payment_data': {
            '_name': 'payment_data',
            'payment_method': {
                '_name': 'payment_method',
                '_value': invoice.payment_type
            },
            'creditor': _construct_creditor_debtor_data(invoice.issuer, 'creditor'),
            'debtor': _construct_creditor_debtor_data(invoice.recipient, 'debtor'),
            'execution_time': {
                '_name': 'requested_execution_date',
                '_value': invoice.date_issued.strftime('%Y-%m-%d')
            },
            'amount': {
                '_name': 'amount',
                '_value': str(invoice.total_with_tax)
            },
            'currency': {
                '_name': 'currency',
                '_value': invoice.currency
            },
            'remittance_information': {
                '_name': 'remittance_information',
                'creditor_reference': {
                    '_name': 'creditor_structured_reference',
                    '_value': invoice.payment_reference
                },
                'additional_info': {
                    '_name': 'additional_remittance_information',
                    '_value': invoice.additional_remittance_information
                }
            },
            'purpose': {
                '_name': 'purpose',
                '_value': invoice.payment_purpose
            }
        },
        'attachments': {
            '_name': 'attachments',
            'hash': {
                '_name': 'hash',
                '_value': '0000000000000000000000000000000000000000',
            },
            'size': {
                '_name': 'size',
                '_value': '0'
            },
            'count': {
                '_name': 'count',
                '_value': str(len(attachments))
            }
        }
    }

    for i, attachment in enumerate(attachments):
        envelope['attachments'][f'attachment_{i}'] = {
            '_name': 'attachment',
            'filename': {
                '_name': 'filename',
                '_value': attachment[0]
            },
            'size': {
                '_name': 'size',
                '_value': str(attachment[2])
            },
            'type': {
                '_name': 'type',
                '_value': attachment[1]
            }
        }

    return envelope


def _construct_sender_receiver_data(entity, entity_type='sender', bic=None):
    data = {
        '_name': entity_type,
        'name': {
            '_name': 'name',
            '_value': entity.name[:70]  # max length
        },
        'country': {
            '_name': 'country',
            '_value': entity.country_iso_code
        },
        'address': {
            '_name': 'address',
            '_value': entity.address[:70]
        },
        'zip_code': {
            '_name': 'address',
            '_value': str(entity.zip_code)
        },
        'vat_id': {
            '_name': f'{entity_type}_identifier',
            '_value': str(entity.vat_id)
        },
        'e_address': {
            '_name': f'{entity_type}_eddress',
            'bic': {
                '_name': f'{entity_type}_agent',
                '_value': bic if bic else entity.bic
            },
            'iban': {
                '_name': f'{entity_type}_mailbox',
                '_value': entity.iban
            }
        }
    }

    return data


def _construct_creditor_debtor_data(entity,  entity_type='creditor'):
    data = {
        '_name': entity_type,
        'name': {
            '_name': 'name',
            '_value': entity.name[:70]  # max length
        },
        'country': {
            '_name': 'country',
            '_value': entity.country_iso_code
        },
        'address': {
            '_name': 'address',
            '_value': entity.address[:70]
        },
        'zip_code': {
            '_name': 'address',
            '_value': str(entity.zip_code)
        },
        'bic': {
            '_name': f'{entity_type}_agent',
            '_value': entity.bic
        },
        'iban': {
            '_name': f'{entity_type}_account',
            '_value': entity.iban
        }
    }

    return data



import textwrap


def construct_invoice_json(invoice):
    NS = 'http://www.w3.org/2001/XMLSchema-instance'

    data = {
        '_name': 'IzdaniRacunEnostavni',
        '_attrs': [("{%s}noNamespaceSchemaLocation" % NS, 'http://www.gzs.si/e-poslovanje/sheme/eSLOG_1-6_EnostavniRacun.xsd')],
        '_ns': {
            'ds': 'http://www.w3.org/2000/09/xmldsig#',
            'xsd': 'http://uri.etsi.org/01903/v1.1.1#',
            'xsi': NS
        },
        'invoice': {
            '_name': 'Racun',
            '_attrs': [('Id', 'data')],
            'header': construct_header_data(invoice),
            'date_issued': construct_date_data(invoice.date_issued_code, invoice.date_issued),
            'date_of_service': construct_date_data(invoice.date_of_service_code, invoice.date_of_service),
            'currency': construct_currency_data(invoice.currency),
            'location': construct_location_data(invoice.location_code, invoice.location_address),
            'issuer': construct_company_data(invoice.issuer, 'II'),
            'buyer': construct_company_data(invoice.recipient, 'BY'),
            'recipient': construct_company_data(invoice.recipient, 'IV'),
            'payment_terms': construct_payment_terms_data(invoice.date_due_code, invoice.date_due),
            'reference_data': construct_reference_data(invoice.total_with_tax, invoice.payment_reference),

        }  # end invoice
    }

    for i, reference_document in enumerate(invoice.reference_documents):
        data['invoice'][f"reference_document_{i}"] = construct_reference_document_data(reference_document)

    if invoice.global_discount_amount:
        data['invoice']['global_discount'] = construct_global_discount_data(invoice.global_discount_amount, invoice.global_discount_percentage)

    if invoice.intro_text:
        data['invoice']['intro_text'] = construct_custom_text_data('AAI', 'GLAVA_TEKST', invoice.intro_text)

    for i, item in enumerate(invoice.document_items):
        data['invoice'][f"item_{i}"] = construct_item_data(item)

    for i, ts in enumerate(invoice.tax_summaries):
        data['invoice'][f"tax_summary_{i}"] = construct_tax_summary_data(ts)

    # add final sums to invoice
    # Total without discount
    data['invoice']['sums_without_discounts'] = construct_sums_data(amount=invoice.total_without_discount, sum_type='79')
    # Discounts amount
    data['invoice']['sums_discounts'] = construct_sums_data(amount=invoice.total_without_discount - invoice.total_without_tax, sum_type='53')
    # Tax base sums
    data['invoice']['sums_tax_base_amount'] = construct_sums_data(amount=invoice.total_without_tax, sum_type='125')
    # Taxes amount
    data['invoice']['sums_taxes'] = construct_sums_data(amount=invoice.total_with_tax - invoice.total_without_tax, sum_type='176')
    # Total amount - with taxes
    data['invoice']['sums_total_amount'] = construct_sums_data(amount=invoice.total_with_tax, sum_type='86')

    if invoice.outro_text:
        data['invoice']['outro_text'] = construct_custom_text_data('AAI', 'DODATNI_TEKST', invoice.outro_text)

    return data


def construct_header_data(invoice):
    header = {
        '_name': 'GlavaRacuna',
        'invoice_type': {
            '_name': 'VrstaRacuna',
            '_value': invoice.invoice_type,
        },
        'invoice_number': {
            '_name': 'StevilkaRacuna',
            '_value': invoice.invoice_number,
        },
        'invoice_function': {
            '_name': 'FunkcijaRacuna',
            '_value': invoice.invoice_function,
        },
        'payment_type': {
            '_name': 'NacinPlacila',
            '_value': invoice.payment_type,
        },
        'payment_purpose': {
            '_name': 'KodaNamena',
            '_value': invoice.payment_purpose
        }
    }

    return header


def construct_date_data(date_code, date):
    date = {
        '_name': 'DatumiRacuna',
        'date_type': {
            '_name': 'VrstaDatuma',
            '_value': date_code
        },
        'date': {
            '_name': 'DatumRacuna',
            '_value': date.isoformat()
        }
    }

    return date


def construct_currency_data(currency):
    currency = {
        '_name': 'Valuta',
        'currency_type': {
            '_name': 'VrstaValuteRacuna',
            '_value': '2',
        },
        'currency_code': {
            '_name': 'KodaValute',
            '_value': currency,
        },
    }

    return currency


def construct_location_data(location_code, location_address):
    location = {
        '_name': 'Lokacije',
        'location_code': {
            '_name': 'VrstaLokacije',
            '_value': location_code,
        },
        'location_name': {
            '_name': 'NazivLokacije',
            '_value': location_address,
        },
    }

    return location


def construct_company_data(business, business_type='II'):
    data = {
        '_name': 'PodatkiPodjetja',
        'info': {
            '_name': 'NazivNaslovPodjetja',
            'type': {
                '_name': 'VrstaPartnerja',
                '_value': business_type
            },
            'name': {
                '_name': 'NazivPartnerja',
            },
            'address': {
                '_name': 'Ulica',
            },
            'city': {
                '_name': 'Kraj',
                '_value': business.city
            },
            'country': {
                '_name': 'NazivDrzave',
                '_value': business.country
            },
            'zip_code': {
                '_name': 'PostnaStevilka',
                '_value': str(business.zip_code)
            },
            'country_code': {
                '_name': 'KodaDrzave',
                '_value': business.country_iso_code
            },
        }
    }

    # Add business name
    business_name_split = textwrap.wrap(business.name, 35, break_long_words=True)

    for i, bn_part in enumerate(business_name_split):
        i = i + 1  # Start from 1
        data['info']['name'][f"part_{i}"] = {
            '_name': f"NazivPartnerja{i}",
            '_value': bn_part
        }

        if i == 4:
            break  # Stop at max length

    # Add business name
    addr_split = textwrap.wrap(business.address, 35, break_long_words=True)

    for i, addr_part in enumerate(addr_split):
        i = i + 1  # Start from 1
        data['info']['address'][f"part_{i}"] = {
            '_name': f"Ulica{i}",
            '_value': addr_part
        }

        if i == 4:
            break  # Stop at max length

    if business.iban:
        data['financial_info'] = {
            '_name': 'FinancniPodatkiPodjetja',
            'bank_account_info': {
                '_name': 'BancniRacun',
                'iban': {
                    '_name': 'StevilkaBancnegaRacuna',
                    '_value': business.iban
                },
                'bic': {
                    '_name': 'BIC',
                    '_value': business.bic
                }
            }
        }

    if business.vat_id:
        data['vat_id'] = {
            '_name': 'ReferencniPodatkiPodjetja',
            'type': {
                '_name': 'VrstaPodatkaPodjetja',
                '_value': 'VA'
            },
            'vat': {
                '_name': 'PodatekPodjetja',
                '_value': str(business.vat_id)
            }
        }

    if business.registration_number:
        data['registration_number'] = {
            '_name': 'ReferencniPodatkiPodjetja',
            'type': {
                '_name': 'VrstaPodatkaPodjetja',
                '_value': 'GN'
            },
            'vat': {
                '_name': 'PodatekPodjetja',
                '_value': str(business.registration_number)
            }
        }

    return data


def construct_payment_terms_data(date_due_code, date_due):
    payment_terms = {
        '_name': 'PlacilniPogoji',
        'term_data': {
            '_name': 'PodatkiORokih',
            'term_code': {
                '_name': 'VrstaPogoja',
                '_value': '3',
            }
        },
        'term_due': {
            '_name': 'PlacilniRoki',
            'term_due_type': {
                '_name': 'VrstaDatumaPlacilnegaRoka',
                '_value': date_due_code,
            },
            'term_due_date': {
                '_name': 'Datum',
                '_value': date_due.isoformat(),
            },
        }
    }

    return payment_terms


def construct_reference_data(total_with_tax, payment_reference):
    reference = {
        '_name': 'PovzetekZneskovRacuna',
        'invoice_amounts': {
            '_name': 'ZneskiRacuna',
            'type': {
                '_name': 'VrstaZneska',
                '_value': '9'  # Amount to be paid
            },
            'amount': {
                '_name': 'ZnesekRacuna',
                '_value': str(total_with_tax)
            }
        },
        'reference': {
            '_name': 'SklicZaPlacilo',
            'ref_type': {
                '_name': 'SklicPlacila',
                '_value': 'PQ'
            },
            'ref_number': {
                '_name': 'StevilkaSklica',
                '_value': payment_reference
            }
        }
    }

    return reference


def construct_reference_document_data(reference_document):
    reference_doc_data = {
        '_name': 'ReferencniDokumenti',
        '_attrs': [('VrstaDokumenta', reference_document.type_code)],
        'document_number': {
            '_name': "StevilkaDokumenta",
            '_value': reference_document.document_number
        }
    }

    return reference_doc_data


def construct_global_discount_data(discount_amount, discount_percentage):
    discount_data = {
        '_name': 'GlobalniPopusti',
        'description': {
            '_name': 'OpisPopusta',
            '_value': 'SKUPNI POPUST',
        },
        'type': {
            '_name': 'TipPopusta',
            '_value': 'PP',
        },
        'percentage': {
            '_name': 'OdstotekPopusta',
            '_value': str(discount_percentage)
        },
        'amount': {
            '_name': 'ZnesekPopusta',
            '_value': str(discount_amount)
        },
    }

    return discount_data


def construct_custom_text_data(text_format, text_type, text):
    """
    Text must be split into 70 chars long strings.
    :param text_format: AAI or other predefined formats
    :param text_type:
    :param text:
    :return:
    """
    text_split = textwrap.wrap(text, 70, break_long_words=True)

    custom_text = {
        '_name': 'PoljubnoBesedilo',
        'format': {
            '_name': 'VrstaBesedila',
            '_value': text_format,
        },
        'content': {
            '_name': 'Besedilo',
            'text_1': {
                '_name': 'Tekst1',
                '_value': text_type,
            },
        }
    }

    for i, txt in enumerate(text_split):
        # Since text_1 is used for text_type we must enumerate from 2 onwards
        i = i + 2

        custom_text['content'][f"text_{i}"] = {
            '_name': f"Tekst{i}",
            '_value': txt
        }

        # Stop the loop if index is 5 - we can't place any more than  that in XML.
        if i == 5:
            break

    return custom_text


def construct_item_data(item):
    data = {
        '_name': 'PostavkeRacuna',
        'info': {
            '_name': 'Postavka',
            'row_num': {
                '_name': 'StevilkaVrstice',
                '_value': str(item.row_number)
            }
        },
        'description': {
            '_name': 'OpisiArtiklov',
            'code': {
                '_name': 'KodaOpisaArtikla',
                '_value': 'F',
            },
            'name': {
                '_name': 'OpisArtikla',
                'desc_1': {
                    '_name': 'OpisArtikla1',
                    '_value': item.item_name[:35]  # Only 35 chars...
                }
            }
        },
        'quantity': {
            '_name': 'KolicinaArtikla',
            'qty_type': {
                '_name': 'VrstaKolicine',
                '_value': item.quantity_type,
            },
            'qty': {
                '_name': 'Kolicina',
                '_value': str(item.quantity)
            },
            'unit': {
                '_name': 'EnotaMere',
                '_value': item.unit
            }
        },
        'value_before_discount': {
            '_name': 'ZneskiPostavke',
            'type': {
                '_name': 'VrstaZneskaPostavke',
                '_value': '203'  # Total before discount
            },
            'amount': {
                '_name': 'ZnesekPostavke',
                '_value': "%.2f" % (item.price_without_tax * item.quantity)
            }
        },
        'value_total': {
            '_name': 'ZneskiPostavke',
            'type': {
                '_name': 'VrstaZneskaPostavke',
                '_value': '38'  # Total with discount
            },
            'amount': {
                '_name': 'ZnesekPostavke',
                '_value': str(item.total_with_tax)
            }
        },
        'price': {
            '_name': 'CenaPostavke',
            'value': {
                '_name': 'Cena',
                '_value': str(item.price_without_tax)
            }
        },
        'tax_info': {
            '_name': 'DavkiPostavke',
            'taxes': {
                '_name': 'DavkiNaPostavki',
                'type': {
                    '_name': 'VrstaDavkaPostavke',
                    '_value': 'VAT'
                },
                'vat_percentage': {
                    '_name': 'OdstotekDavkaPostavke',
                    '_value': str(item.tax_rate)
                }
            },
            'tax_amounts_base': {
                '_name': 'ZneskiDavkovPostavke',
                'type': {
                    '_name': 'VrstaZneskaDavkaPostavke',
                    '_value': '125'
                },
                'amount': {
                    '_name': 'Znesek',
                    '_value': str(item.total_without_tax)
                }
            },
            'tax_amounts_tax': {
                '_name': 'ZneskiDavkovPostavke',
                'type': {
                    '_name': 'VrstaZneskaDavkaPostavke',
                    '_value': '124'
                },
                'amount': {
                    '_name': 'Znesek',
                    '_value': str(item.total_with_tax - item.total_without_tax)
                }
            }
        }
    }

    if item.discount_percentage:
        data['discount'] = {
            '_name': 'OdstotkiPostavk',
            'identification': {
                '_name': 'Identifikator',
                '_value': 'A',  # Discount
            },
            'type': {
                '_name': 'VrstaOdstotkaPostavke',
                '_value': '12',  # Discount
            },
            'percentage': {
                '_name': 'OdstotekPostavke',
                '_value': str(item.discount_percentage)
            },
            'type_amount': {
                '_name': 'VrstaZneskaOdstotka',
                '_value': '204'
            },
            'amount': {
                '_name': 'ZnesekOdstotka',
                '_value': str(item.discount_amount)
            }

        }

    return data


def construct_tax_summary_data(tax_summary):
    data = {
        '_name': 'PovzetekDavkovRacuna',
        'summary': {
            '_name': 'DavkiRacuna',
            'tax_type': {
                '_name': 'VrstaDavka',
                '_value': 'VAT'
            },
            'tax_percentage': {
                '_name': 'OdstotekDavka',
                '_value': str(tax_summary.tax_rate)
            },
        },
        'amount_base': {
            '_name': 'ZneskiDavkov',
            'type': {
                '_name': 'VrstaZneskaDavka',
                '_value': '125'  # Osnova
            },
            'amount': {
                '_name': 'ZnesekDavka',
                '_value': str(tax_summary.tax_base)
            }
        },
        'amount_tax': {
            '_name': 'ZneskiDavkov',
            'type': {
                '_name': 'VrstaZneskaDavka',
                '_value': '124'  # Tax amount
            },
            'amount': {
                '_name': 'ZnesekDavka',
                '_value': str(tax_summary.tax_amount)
            }
        }
    }

    return data


def construct_sums_data(amount, sum_type, ref=None):
    data = {
        '_name': 'PovzetekZneskovRacuna',
        'amounts': {
            '_name': 'ZneskiRacuna',
            'type': {
                '_name': 'VrstaZneska',
                '_value': str(sum_type)
            },
            'amount': {
                '_name': 'ZnesekRacuna',
                '_value': "%.2f" % amount
            }
        },
        'ref': {
            '_name': 'SklicZaPlacilo',
            'ref_type': {
                '_name': 'SklicPlacila',
                '_value': 'PQ'
            }
        }
    }

    if ref is not None:
        data['ref']['ref_number'] = {
            '_name': 'StevilkaSklica',
            '_value': ref
        }

    return data



import hashlib

from base64 import b64encode
from datetime import datetime, timezone

from cryptography import x509
from cryptography.hazmat.backends import default_backend

from signxml import XMLSigner

from eracun_generator.builder import build_xml


def ds_tag(tag):
    return "{http://www.w3.org/2000/09/xmldsig#}" + tag


def sign_invoice(invoice_json, key, cert):
    invoice_json = add_temp_sign_data(key, cert, invoice_json)

    xml_content = build_xml(invoice_json)

    signed_xml = XMLSigner(signature_algorithm='rsa-sha1',
                           digest_algorithm='sha1',
                           c14n_algorithm='http://www.w3.org/TR/2001/REC-xml-c14n-20010315').sign(xml_content,
                                                                                                  key=key,
                                                                                                  cert=cert,
                                                                                                  reference_uri=['data','signprops'])

    ds_object = signed_xml[1][0]

    signed_xml[1].remove(ds_object)
    signed_xml[1].append(ds_object)

    signed_xml[1].attrib['Id'] = 'signature'

    return signed_xml


def add_temp_sign_data(key, cert, invoice_json, signed_props_id='signprops'):
    pm = cert
    pem_cert = x509.load_pem_x509_certificate(pm, default_backend())

    digest = hashlib.sha1()
    digest.update(pm)
    digest = b64encode(digest.digest()).decode()

    invoice_json['signature_placeholder'] = {
        '_name': ds_tag('Signature'),
        '_attrs': [('Id', 'placeholder')]
    }

    invoice_json['signature_placeholder']['sign_temp_data'] = {
        '_name': ds_tag('Object'),
        'qualifying_props': {
            '_name': 'QualifyingProperties',
            '_attrs': [('Target', '#signature')],
            'signed_props': {
                '_name': 'SignedProperties',
                '_attrs': [('Id', signed_props_id)],
                'sign_props': {
                    '_name': 'SignedSignatureProperties',
                    'time': {
                        '_name': 'SigningTime',
                        '_value': datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
                    },
                    'signing_cert': {
                        '_name': 'SigningCertificate',
                        'cert': {
                            '_name': 'Cert',
                            'digest': {
                                '_name': 'CertDigest',
                                'method': {
                                    '_name': 'DigestMethod',
                                    '_attrs': [('Algorithm', 'http://www.w3.org/2000/09/xmldsig#sha1')],
                                },
                                'value': {
                                    '_name': 'DigestValue',
                                    '_value': digest
                                }
                            },
                            'issuer_serial': {
                                '_name': 'IssuerSerial',
                                'name': {
                                    '_name': 'X509IssuerName',
                                   # '_attrs': [('xmlns', 'http://www.w3.org/2000/09/xmldsig#')],
                                    '_value': pem_cert.subject.rfc4514_string()
                                },
                                'serial': {
                                    '_name': 'X509SerialNumber',
                                   # '_attrs': [('xmlns', 'http://www.w3.org/2000/09/xmldsig#')],
                                    '_value': str(pem_cert.serial_number)
                                }
                            }
                        },
                    },
                    'policy': {
                        '_name': 'SignaturePolicyIdentifier',
                        'implied': {
                            '_name': 'SignaturePolicyImplied'
                        }
                    }
                }
            }
        }
    }

    return invoice_json



from setuptools import setup, find_packages

setup(
    name='eracun_generator',
    packages=['eracun_generator', 'eracun_generator.envelope'],
    version='0.2.3',
    description='e-SLOG e-Racun v1.6.1 and v2.0 XML generator',
    author='Boris Savic',
    author_email='boris70@gmail.com',
    url='https://github.com/boris-savic/python-eracun',
    download_url='https://github.com/boris-savic/python-eracun/tarball/0.2.3',
    keywords=['python eracun', 'e-racun', 'e-slog'],
    classifiers=[],
    install_requires=[
        'lxml>=4.4.1',
        'signxml>=2.6.0'
    ]
)



from decimal import Decimal

from lxml import etree

from eracun_generator.builder import build_xml
from eracun_generator.definitions import construct_invoice_json
from eracun_generator.definitionsV2 import construct_invoice_json as construct_invoice_jsonV2
from eracun_generator.envelope.utils import convert_invoice_to_envelope
from eracun_generator.utils import ds_tag, sign_invoice


class Business:
    """
    Prepare Business object for Issuer and Recipient instances.
    """
    def __init__(self, name, address, city, zip_code, country, country_iso_code, vat_id, iban=None, bic=None, registration_number=None):
        self.name = name
        self.address = address
        self.city = city
        self.zip_code = zip_code
        self.country = country
        self.country_iso_code = country_iso_code
        self.iban = iban.replace(" ", "") if iban else None  # remove any spaces
        self.bic = bic.ljust(11, 'X') if bic else None  # Schema requires 11 chars long - add Xs at the end of BIC/SWIFT
        self.registration_number = registration_number
        self.vat_id = vat_id


class ReferenceDocument:
    TYPE_PROFORMA_INVOICE = 'AAB'
    TYPE_DELIVERY_ORDER_NUMBER = 'AAJ'
    TYPE_DELIVERY_FORM = 'AAK'
    TYPE_BENEFICIARYS_REFERENCE = 'AFO'
    TYPE_CONSODILDATED_INVOICE = 'AIZ'
    TYPE_MESSAGE_BATCH_NUMBER = 'ALL'
    TYPE_RECEIVING_ADVICE_NUMBER = 'ALO'
    TYPE_COMMERCIAL_ACCOUNT_SUMMARY = 'APQ'
    TYPE_CREDIT_NOTE = 'CD'
    TYPE_CUSTOMER_REFERENCE_NUMBER = 'CR'
    TYPE_CONTRACT = 'CT'
    TYPE_DEBIT_NOTE = 'DL'
    TYPE_DELIVERY_NOTE = 'DQ'
    TYPE_IMPORT_LICENCE_NUMBER = 'IP'
    TYPE_INVOICE = 'IV'
    TYPE_ORDER_NUMBER = 'ON'
    TYPE_PRICE_LIST_NUMBER = 'PL'
    TYPE_PURCHASE_ORDER_RESPONSE_NUMBER = 'POR'
    TYPE_EXPORT_REFERENCE_NUMBER = 'RF'
    TYPE_SPECIFICATION_NUMBER = 'SZ'
    TYPE_ORDER_NUMBER_SUPPLIER = 'VN'

    def __init__(self, document_number, type_code=TYPE_ORDER_NUMBER):
        self.document_number = document_number
        self.type_code = type_code


class InvoiceItem:
    def __init__(self,
                 row_number,
                 item_name,
                 quantity,
                 price_without_tax,
                 total_with_tax,
                 total_without_tax,
                 tax_rate,
                 tax_rate_type='S',
                 discount_percentage=None,
                 discount_amount=None,
                 unit='PCE',
                 ean=''):
        self.row_number = row_number
        self.item_name = item_name
        self.quantity = quantity
        self.price_without_tax = price_without_tax   # CenaPostavke

        self.total_with_tax = total_with_tax         # total - after discount and tax
        self.total_without_tax = total_without_tax   # total - after discount

        self.discount_percentage = discount_percentage
        self.discount_amount = discount_amount

        self.tax_rate = tax_rate
        self.tax_rate_type = tax_rate_type

        self.unit = unit
        self.ean = ean

        self.item_description_code = 'F'
        self.quantity_type = '47'


class TaxSummary:
    def __init__(self, tax_rate, tax_amount, tax_base, tax_type='S'):
        self.tax_rate = tax_rate
        self.tax_amount = tax_amount
        self.tax_base = tax_base
        self.tax_type = tax_type


class Invoice:
    TYPE_MEASURED_SERVICES = '82'
    TYPE_FINANCIAL_CREDIT_NOTE = '83'
    TYPE_FINANCIAL_DEBIT_NOTE = '84'
    TYPE_INVOICING_DATA_SHEET = '130'
    TYPE_PROFORMA_INVOICE = '325'
    TYPE_INVOICE = '380'
    TYPE_CREDIT_NOTE = '381'
    TYPE_COMMISION_NOTE = '382'
    TYPE_DEBIT_NOTE = '383'
    TYPE_CORRECTED_INVOICE = '384'
    TYPE_CONSOLIDATED_INVOICE = '385'
    TYPE_PREPAYMENT_INVOICE = '386'
    TYPE_SELF_BILLED_INVOICE = '389'
    TYPE_DELCREDRE_INVOICE = '390'
    TYPE_FACTORED_INVOICE = '393'

    FUNCTION_CANCELLATION = '1'
    FUNCTION_REPLACE = '5'
    FUNCTION_DUPLICATE = '7'
    FUNCTION_ORIGINAL = '9'
    FUNCTION_COPY = '31'
    FUNCTION_ADDITIONAL_TRANSMISSION = '43'

    PAYMENT_REQUIRED = '0'
    PAYMENT_DIRECT_DEBIT = '1'
    PAYMENT_ALREADY_PAID = '2'
    PAYMENT_OTHER_NO_PAYMENT = '3'

    LOCATION_PAYMENT = '57'
    LOCATION_ISSUED = '91'
    LOCATION_SALE = '162'

    def __init__(self,
                 issuer,  # Business object
                 recipient,  # Business object
                 invoice_number,
                 total_without_tax,
                 total_with_tax,
                 location_address,
                 date_issued,
                 date_of_service,
                 date_due,
                 payment_reference,
                 global_discount_amount=None,
                 global_discount_percentage=None,
                 intro_text=None,   # type AAI = general information text  - GLAVA_TEXT
                 outro_text=None,   # type AAI = general information text  - DODATNI_TEKST
                 invoice_type=TYPE_INVOICE,
                 currency='EUR',
                 invoice_function=FUNCTION_ORIGINAL,
                 payment_type=PAYMENT_REQUIRED,
                 payment_purpose="GDSV",
                 additional_remittance_information=None,
                 location_code=LOCATION_ISSUED):
        # Business objects
        self.issuer = issuer
        self.recipient = recipient

        self.invoice_number = invoice_number

        self.total_without_tax = total_without_tax
        self.total_with_tax = total_with_tax

        self.location_address = location_address

        self.date_issued = date_issued
        self.date_of_service = date_of_service
        self.date_due = date_due

        self.payment_reference = payment_reference

        self.global_discount_amount = global_discount_amount
        self.global_discount_percentage = global_discount_percentage

        self.intro_text = intro_text
        self.outro_text = outro_text

        self.invoice_type = invoice_type
        self.currency = currency
        self.invoice_function = invoice_function
        self.payment_type = payment_type
        self.payment_purpose = payment_purpose
        self.additional_remittance_information = additional_remittance_information or f'Racun st. {self.invoice_number}'

        self.location_code = location_code

        self.date_issued_code = '137'
        self.date_of_service_code = '35'
        self.date_due_code = '13'

        self.reference_documents = []

        self.document_items = []

        self.total_without_discount = Decimal('0')

        self.tax_summaries = []

    def add_reference_document(self, reference_document_number, reference_type=ReferenceDocument.TYPE_ORDER_NUMBER):
        self.reference_documents.append(ReferenceDocument(reference_document_number, reference_type))

    def add_tax_summary(self, tax_rate, tax_base, tax_amount):
        self.tax_summaries.append(TaxSummary(tax_rate=tax_rate, tax_amount=tax_amount, tax_base=tax_base))

    def add_item(self,
                 row_number,
                 item_name,
                 quantity,
                 price_without_tax,
                 total_with_tax,
                 total_without_tax,
                 tax_rate,
                 ean=None,
                 discount_percentage=None,
                 discount_amount=None,
                 unit='PCE'):
        self.document_items.append(InvoiceItem(row_number=row_number,
                                               item_name=item_name,
                                               quantity=quantity,
                                               price_without_tax=price_without_tax,
                                               total_with_tax=total_with_tax,
                                               total_without_tax=total_without_tax,
                                               tax_rate=tax_rate,
                                               ean=ean,
                                               discount_percentage=discount_percentage,
                                               discount_amount=discount_amount,
                                               unit=unit))

        self.total_without_discount = self.total_without_discount + (quantity * price_without_tax)

    def render_envelope(self, attachments=None, sender_bic=None, recipient_bic=None):
        """
        Render envelope to be included with the eRacun.

        Every Envelope will have eRacun.xml as an attachment. It is recommended to
        include PDF invoice as an attachment as well.

        To include attachment data in envelope pass in the attachment name and format. For example

        attachments=[('invoice.pdf', 'PDF')]

        :param attachments:
        :param sender_bic: Set to UJPLSI20ICL for DEV environment and UJPLSI2DICL for production environment if
                           the issuer is government budget user.
        :param recipient_bic: Set to UJPLSI20ICL for DEV environment and UJPLSI2DICL for production environment if
                              the recipient is government budget user.
        :return:
        """
        if attachments is None:
            attachments = []

        attachments.insert(0, ('eRacun.xml', 'XML', 0))

        return ("%s%s" % ('<?xml version="1.0" encoding="UTF-8"?>\n',
                          etree.tostring(build_xml(convert_invoice_to_envelope(self, attachments, sender_bic, recipient_bic)),
                                         pretty_print=True,
                                         xml_declaration=False,
                                         encoding="utf-8").decode('utf-8')))

    def render_xml(self, key=None, cert=None, v2=True):
        if v2:
            xml_content = build_xml(construct_invoice_jsonV2(self))

            if key and cert:
                xml_content = sign_invoice(construct_invoice_jsonV2(self), key, cert)

            return ("%s%s" % ('<?xml version="1.0" encoding="UTF-8"?>\n',
                            etree.tostring(xml_content,
                                            pretty_print=False,
                                            xml_declaration=False,
                                            method="c14n",
                                            ).decode('utf-8')))
        else :
            xml_content = build_xml(construct_invoice_json(self))

            if key and cert:
                xml_content = sign_invoice(construct_invoice_json(self), key, cert)

            return ("%s%s" % ('<?xml version="1.0" encoding="UTF-8"?>\n',
                            etree.tostring(xml_content,
                                            pretty_print=False,
                                            xml_declaration=False,
                                            method="c14n",
                                            ).decode('utf-8')))



from lxml import etree


def build_xml(data):
    if data.get('_ns'):
        tag = etree.Element(data['_name'], nsmap=data['_ns'])
    else:
        tag = etree.Element(data['_name'])

    for attr, val in data.get('_attrs', []):
        tag.attrib[attr] = val

    for child, child_data in data.items():
        if not child.startswith('_'):
            child_tag = build_xml(child_data)
            tag.append(child_tag)

    if data.get('_value', None) is not None:
        tag.text = data['_value']

    # Sort the child elements
    if '_sorting' in data:
        tag[:] = sorted(tag, key=lambda element: data['_sorting'].index(element.tag))

    return tag
